#include "StdAfx.h"
#include "simpleworldeffect.h"
#include "Core\HField.h"
#include "Core\PacketVisitor.h"
#include "Core\WMException.h"

// Bring in the WM namespace so we can work with it easier.
using namespace WorldMachine;

Device *SimpleWorldEffectMaker() { return new SimpleWorldEffect; };
void SimpleWorldEffectKiller(Device *spr) { delete spr; };

SimpleWorldEffect::SimpleWorldEffect(void)
{
	lifeptrs.maker = SimpleWorldEffectMaker;
	lifeptrs.killer = SimpleWorldEffectKiller;
	strncpy(lifeptrs.nametag, "EXSW", 4);
	SetLinks(0, 1);	// generators have 0 inputs by default. we'll just re-emphasize that..

	AddParam(Parameter("Spacing", 1.0f, 0.1f, 10.0f));
}

#define SWE_SPACING		ParmFRef("Spacing")

SimpleWorldEffect::~SimpleWorldEffect(void)
{
}

// You are REQUIRED to write out and read in your 4-byte Device ID tag, as shown below. It is also recommended that your device
// have an internal version number that you keep track of to allow backwards-compatibility if you ever change your Device save layout...
// Also, make sure you call your super class's Load and Save functions as shown here. Failure to do so will cause major problems.
bool SimpleWorldEffect::Load(std::istream &in, LoadContext &context) {
	char tag[5];
	in.read( tag, 4);
	if (strncmp(tag, lifeptrs.nametag, 4) == 0) {
		int ver = 0;
		in.read((char*) &ver, 1);
		return Generator::Load(in, context);
	}
	else
		return false;
};

bool SimpleWorldEffect::Save(std::ostream &out, SaveContext &context) {
	out.write( lifeptrs.nametag, 4);
	int ver = 1;
	out.write((char*) &ver, 1);
	return Generator::Save(out, context);
};

bool SimpleWorldEffect::Activate(BuildContext &context) {
// Create a new heightfield using current specified world size
	HFPointer map = GetNewHF(context);

	float spacing = SWE_SPACING;

// get the current origin, modified by the scaling mode and any coordinate offset 
//   supplied on the coordinate input to the device. The origin is in WORLD SPACE.
	CoordF origin = getOrigin();	

// Iterate over the entire heightfield, supplying a value at each location.
// We're going to make a sine-type pattern based on the distance to the origin.
// We want it to produce the correct result no matter the aspect ratio, scaling, or panning of the heightfield.



// There are several ways to do this, including using a function pair in BuildContext that can take coordinates to/from worldspace.

// Here we'll demonstrate the use of a PacketVisitor class and a C++11 lambda to make our life MUCH easier.
// Not just easier, but faster: the PacketVisitor automatically uses uses pointer arithmetic and accumulators for you, to give you a nice iteration speed.
// Check the docs for more details, but it is this simple:

// Create the packet visitor on our heightfield. You can work with many (same-sized) grids at the same time by calling with multiple arguments, or calling add(..).
	PacketVisitor<HField> visitor(*map);

	// Visit each grid spot in no particular order.
	// Note that we're using a C++11 style lambda, you could also define a function object for this.
	visitor.visitAll( [=] (PacketVisitor<HField>::Location const& loc) {
		// Get the distance from the current grid pixel to the origin. We find the world space location of the current grid pixel by simply retrieving it from the location structure passed to our lambda.
		CoordF delta = loc.wloc - origin;

		// distance in worldspace from this height pixel to the origin
		float distance = delta.length();

		// do the cosine pattern, and rescale the result from -1..1 to 0..1.
		float result_value = (0.5f * cos(3.14159f * distance * spacing)) + 0.5f;

		// assign the results to packet #0 in our visitor, at the current location.
		loc[0] = result_value;
	});

// Handoff to output(s)
	StoreData(map,0, context); // pass the heightfield to our output stage.
// Success!
	return true;
}
