#pragma once

#include "AreaPacket.h"

/** Abstract templated base class derived from WMAreaPacket that not just covers some area,
	but does so with a regular grid. This is thus the immediate ancestor of HField and BMPacket.

	You can create all kinds of new grid-based packet types by simply inheriting from this class, ie:

	struct myDataType { .. };
	class MyGridPacket : public WMGridPacket<myDataType> { .. };

	If you do this, make sure your application core knows how to construct a packet of the new class.
*/
template <class T> class WMGridPacket : public WMAreaPacket  {
public:
	WMGridPacket<T> &operator=(const WMGridPacket<T> &other);

	typedef typename T value_type;
	typedef typename T* iterator;
	typedef typename const T* const_iterator;

	virtual bool Load( std::istream &input ) override;
	virtual bool Save( std::ostream &output ) override;

	/// Initialze packet using provided Packetcontext for spatial and resolution information
	virtual void Init(const PacketContext &_size);
	/// Initializes HField and matches others' attributes, but DOES NOT copy data contents! Use copy constructor if you want that.
	virtual void InitFrom(const WMGridPacket<T> &other);
	virtual size_t getHeapUsage() { if (!datagrid) return 0; return areacontext.resarea() * sizeof(T);	};

	/// Clear the grid to a given value
	virtual void ClearTo(T clearval);

	/// Grid datatypes are huge and thus excellent candidates for being paged out when low on memory.
	/// If this for some reason doesnt work for your class, override this in your derived class.
	virtual bool isTypePageable() { return true; };
//	virtual bool queryType(const int type);

/// @name Data Access Methods
/// Basic access methods. These are in order from fastest to slowest methods for accessing heightfield
/// memory.
//@{
	inline T *GetDataPtr();						///< Return a direct pointer to grid memory. USE WITH CAUTION!!!
	inline T &operator[](const size_t index);	///< lowest overhead besides direct pointer access
	inline T value(const Coord loc) const;		///< fastest coordinate-based accessor
	inline T &operator[](const Coord loc);		///< 
	inline T &operator[](const Vector2 loc);	///< Truncates coordinates to an integer before retrieving. If you want bilinear interpolated values, use HField::valueF
	inline T const& operator[](const size_t index) const;
	inline T valueSafe(const Coord loc) const;	///< Does bounds checking (clip-to-nearest) on coordinates so that out-of-range access works (slower)
	inline T valueSafeZero(const Coord loc) const;///< Boundscheck but returns 0.0f for out-of-bound values
	inline T valueF(const CoordF loc) const;	///< Get bilinear-interpolated value. (slow)
	inline T valueSafeF(const CoordF loc) const;///< Interpolated, boundschecked retrieval. (slowest)

	inline T valueUV(const CoordF uv) const;	///< Retrieve a bilinear-interpolated value based upon a UV coordinate (normalized 0..1 range addressing)
//@}

	/// GridLocation is a helper structure that GridPacket will provide to your visitAll() visitor functor.
	struct GridLocation {
		GridLocation(T &_data, Coord loc, CoordF worldloc, size_t _index) : dataref(_data), coord(loc), worldpos(worldloc), index(_index) {};
		T &dataref;			///< Direct reference to the data element at the current location
		Coord coord;		///< grid coordinate of current location
		CoordF worldpos;	///< worldspace coordinate of current location
		size_t index;		///< grid index of current location
	private:
		GridLocation(GridLocation const&other); // not copy assignable, accept by reference!
	};

	/// Quick visitor Function. Visitor will be called with a GridLocation parameter.
	///  visitor(GridLocation &location);
	/// the visitor function may change the referenced data.
	/// This is intentionally simple so that it serves as a drop-in for_each functionality. For more advanced abilities (such as reporting while iterating)
	/// use the PacketVisitor class instead.
	template <typename visitor_type >
	void visitAll( visitor_type visitor);

	size_t getIndex(const Coord &loc);	///< Convert a coordinate into an index value
	size_t getStride();				///< the stride of the grid (normally ==  w() )

	/// Determines the range of values in the packet.
	/// Note that this requires that type T support less-than semantics
	void determineRange(T &min, T &max) const;	

protected:
	WMGridPacket();
	virtual ~WMGridPacket();
	friend class WMPacketFactory;

	virtual void resetContents();

	T *datagrid;
};


// Implementation
#include "GridPacket.cpp.h"
