#pragma once

#include "HField.h"
namespace WorldMachine {

/** Simple wrapper to provide a default value if no heightfield is present.
	Provide a scalar value and a heightfield. If there is no heightfield, it will return the scalar alone.
	Otherwise, it will return the sampled heightfield value.
	@ingroup data
	*/
template < class PacketType, bool bMultiply = false, bool bSafe = false>
class InputWrap
{
public:
	typedef typename PacketType::value_type value_type;
	typedef boost::intrusive_ptr<PacketType> ptrtype;
	InputWrap(boost::intrusive_ptr<PacketType> const& hf, const float default_val = 0.0f) { map = hf; val = default_val; };
	InputWrap(PacketType &hf, const float default_val = 0.0f) { map = &hf; val = default_val; };

	void SetDefault(value_type _val);
	inline value_type value(const Coord &loc) const;
	inline value_type operator[](const Coord &loc) const;
	inline value_type operator[](const size_t index) const;
	inline value_type operator[](const CoordF &loc) const;
protected:
	ptrtype map;
	value_type val;
};

//
// implementation
//

template < class PacketType, bool bMultiply /*= false*/, bool bSafe /*= true*/>
typename InputWrap<typename PacketType, bMultiply, bSafe>::value_type WorldMachine::InputWrap<PacketType, bMultiply, bSafe>::operator[]( const size_t index ) const
{
	if (map) {
		value_type mapval;
		static_assert(!bSafe, "You cannot use the index operator when using Safe Access : we don't know how to correct the bounds. Use a coordinate accessor instead.");
		mapval = (*map)[index];

		if (bMultiply)
			return mapval * val;
		else
			return mapval;
	}
	else
		return val;
}


template < class PacketType, bool bMultiply /*= false*/, bool bSafe /*= true*/>
typename InputWrap<typename PacketType, bMultiply, bSafe>::value_type WorldMachine::InputWrap<PacketType, bMultiply, bSafe>::value( const Coord &loc ) const
{
	if (map) {
		value_type mapval;
		if (bSafe)
			mapval = map->valueSafe(loc);
		else
			mapval = map->value(loc);

		if (bMultiply)
			return mapval * val;
		else
			return mapval;
	}
	else
		return val;
}


template < class PacketType, bool bMultiply /*= false*/, bool bSafe /*= true*/>
typename InputWrap<typename PacketType, bMultiply, bSafe>::value_type WorldMachine::InputWrap<PacketType, bMultiply, bSafe>::operator[]( const Coord &loc ) const
{
	return value(loc);
}


template < class PacketType, bool bMultiply /*= false*/, bool bSafe /*= true*/>
typename InputWrap<typename PacketType, bMultiply, bSafe>::value_type WorldMachine::InputWrap<PacketType, bMultiply, bSafe>::operator[]( const CoordF &loc ) const
{
	if (map) {
		value_type mapval;
		if (bSafe)
			mapval = map->valueSafeF(loc);
		else
			mapval = map->valueF(loc);

		if (bMultiply)
			return mapval * val;
		else
			return mapval;
	}
	else
		return val;
}

template < class PacketType, bool bMultiply /*= false*/, bool bSafe /*= true*/>
void WorldMachine::InputWrap<PacketType, bMultiply, bSafe>::SetDefault( value_type _val )
{
	val = _val;
}

}
