#pragma once

#include "WMPacket.h"
#include "WMAllocator.h"

namespace WorldMachine {

/** A simple container class that contains multiple data packets, and supports serialization. 
 For the most part this is a simple wrapper around a std::vector; additional functionality to come later.
*/
class PacketLibrary
{
public:
	PacketLibrary(bool allow_paged_contents = true);
	~PacketLibrary(void);

	bool Load(std::istream &in);
	bool Save(std::ostream &out);

	/// Number of packets in the library
	size_t size() const;

	/// resize contents to the specified size. 
	/// May invalidate directly held references to PacketPointers inside the library
	/// (but does NOT invalidate the data packets themselves)
	void resize(size_t size);

	/// clear contents
	/// May invalidate directly held references to PacketPointers inside the library
	/// (but does NOT invalidate the data packets themselves)
	void clear();

	/// Returns the tagname of the given packet
	const char*name(size_t index) const;

	/// Set the tagname of the given packet. Only succeeds if a packet exists in that slot
	void setName(size_t index, const char * str);

	/// Add a packet to the library.
	/// May invalidate directly held references to PacketPointers inside the library
	/// (but does NOT invalidate the data packets themselves)
	void addPacket(PacketPointer const&pkt, const char*tag = nullptr);

	// Retrieve packet with index specified. This is a reference so you can use it to modify the contents of the library
	PacketPointer &getPacketReference(size_t index);
	PacketPointer const&getPacketReference(size_t index) const;

	// Operator version of the above. Also returns a ScopedPacketAccess guard so that if you are allowing the contents of the library to be paged for
	// memory reasons, that the packet returned will always be in memory and accessible.
	ScopedPacketAccess operator[](size_t index);
private:
	bool allow_page_contents;
	std::vector<PacketPointer, wmcore_allocator<PacketPointer> > contents;
	static PacketPointer nullref;
};

}
