#pragma once
#include "common.h"
#include "color.h"

// Convenience functions for writing template generic algorithms
namespace WorldMachine {
	namespace datatype {
		// All of these functions allow writing generic code that works with either heightfields or RGB bitmaps

		/// vector of all zeros
		template<class T>
		T zero();

		/// vector of all ones
		template<class T>
		T one();

		/// fill a vector-valued datatype with all the same float value
		template<class T>
		T value(float v);

		// Implementations
		template<>
		inline colorF zero() { colorF c; c.r = c.g = c.b = 0.0f; return c;};
		template<>
		inline float zero() { return 0.0f; };

		template<>
		inline colorF one() { colorF c; c.r = c.g = c.b = 1.0f; return c; };
		template<>
		inline float one() { return 1.0f; };

		template<>
		inline colorF value(float v) { colorF c; c.r = c.g = c.b = v; return c; };
		template<>
		inline float value(float v) { return v; };
		template<>
		inline CoordF value(float v) { return CoordF(v,v); };


		//// Functions

		// Apply a functor to either a scalar or RGB vector valued input
		template<class T>
		class byChannel {
		public:
			template<class Op>
			static T apply(Op applyop, T in1);
			template<class Op>
			static T apply(Op applyop, T in1, T in2);
			template<class Op>
			static T apply(Op applyop, T in1, T in2, T in3);

			template <class Op>
			static float reduce(Op reduceop, T input);

		};

		template<class T> T minval(T in1, T in2);
		template<class T> T maxval(T in1, T in2);

		template<>
		inline float WorldMachine::datatype::maxval<float>( float in1, float in2 )
		{
			return max(in1, in2);
		}

		template<>
		inline colorF WorldMachine::datatype::maxval<colorF>( colorF in1, colorF in2 )
		{
			colorF c;
			c.r = max(in1.r, in2.r);
			c.g = max(in1.g, in2.g);
			c.b = max(in1.b, in2.b);
			return c;
		}

		template<>
		inline float WorldMachine::datatype::minval<float>( float in1, float in2 )
		{
			return min(in1, in2);
		}

		template<>
		inline colorF WorldMachine::datatype::minval<colorF>( colorF in1, colorF in2 )
		{
			colorF c;
			c.r = min(in1.r, in2.r);
			c.g = min(in1.g, in2.g);
			c.b = min(in1.b, in2.b);
			return c;
		}

		template<>
		template<class Op>
		inline float byChannel<float>::apply(Op applyop, float inp1) {
			return applyop(inp1);
		};

		template<>
		template<class Op>
		inline colorF byChannel<colorF>::apply(Op applyop, colorF inp1) {
			return colorF(applyop(inp1.r), applyop(inp1.g), applyop(inp1.b));
		};

		template<>
		template<class Op>
		inline float byChannel<float>::apply(Op applyop, float inp1, float inp2) {
			return applyop(inp1, inp2);
		};

		template<>
		template<class Op>
		inline colorF byChannel<colorF>::apply(Op applyop, colorF inp1, colorF inp2) {
			return colorF(applyop(inp1.r, inp2.r), applyop(inp1.g, inp2.g), applyop(inp1.b, inp2.b));
		};

		template<>
		template<class Op>
		inline float byChannel<float>::apply(Op applyop, float inp1, float inp2, float inp3) {
			return applyop(inp1, inp2, inp3);
		};

		template<>
		template<class Op>
		inline colorF byChannel<colorF>::apply(Op applyop, colorF inp1, colorF inp2, colorF inp3) {
			return colorF(applyop(inp1.r, inp2.r, inp3.r), applyop(inp1.g, inp2.g, inp3.g), applyop(inp1.b, inp2.b, inp3.b));
		};

		template<>
		template <class Op>
		inline float byChannel<colorF>::reduce(Op reduceop, colorF input) {
			return reduceop(input.r) + reduceop(input.g) + reduceop(input.b);
		}

		template<>
		template <class Op>
		inline float byChannel<float>::reduce(Op reduceop, float input) {
			return reduceop(input);
		}

	}
}

/// Specialization of clamp for colors
template <> 
inline void clamp(colorF &value, const colorF min, const colorF max) {
	if (value.r < min.r)
		value.r = min.r;
	else if (value.r > max.r)
		value.r = max.r;

	if (value.g < min.g)
		value.g = min.g;
	else if (value.g > max.g)
		value.g = max.g;

	if (value.b < min.b)
		value.b = min.b;
	else if (value.b > max.b)
		value.b = max.b;
	return;
};